"""
This file defines some common filepaths used across many functions.
"""
import os

# defines top-level paths and current directory for reference
HERE = os.path.abspath(os.path.dirname(__file__))
PROJECT_ROOT = os.path.abspath(os.path.join(HERE, '..'))

# defines secondary paths used misc. throughout project files
TMP_DIR = os.path.abspath(os.path.join(PROJECT_ROOT, "tmp"))
DATA_DIR = os.path.abspath(os.path.join(PROJECT_ROOT, 'data'))
SPY_DATA = os.path.abspath(os.path.join(DATA_DIR, 'SPY.csv'))



from enum import Enum


class TradeAction(Enum):
    """
    Representation of all Trade Actions possible by an Agent.
    """
    BUY: "TradeAction" = 0
    SELL: "TradeAction" = 1
    HOLD: "TradeAction" = 2

    @staticmethod
    def get_action_from_name(name: str) -> "TradeAction":
        """
        Given a string name of an action e.g. Buy, Sell, Hold, return the
        associated trade action.
        """
        if name.lower() == "buy":
            return TradeAction.BUY
        if name.lower() == "sell":
            return TradeAction.SELL
        if name.lower() == "hold":
            return TradeAction.HOLD
        raise Exception(
            f"No TradeAction found for: {name}."
            f"Must be lowercase form of the following: "
            f"{[a.name for a in TradeAction]}")

    @staticmethod
    def get_action_from_value(value: int) -> "TradeAction":
        """
        Given an integer value, map to the associated TradeAction
        """
        for action in TradeAction:
            if action.value == value:
                return action
        raise Exception(
            f"No TradeAction associated with the value: {value}."
            f"Available choices: {[a.value for a in TradeAction]}")



import os
import random
from typing import Tuple, Sequence

import numpy as np
import torch
from torch import tensor

from akademy.common.utils import remove_old_file_versions
from akademy.models.base_models import Agent
from akademy.models.base_models import EpsilonGreedy
from akademy.models.experience_replay_memory import ExperienceReplayMemory
from akademy.models.networks.three_layer_linear import ThreeLayerLinear


class DQNAgent(EpsilonGreedy, Agent):
    """
    DQN Agent that utilizes an experience replay buffer for akademy on past
    data, an epsilon-greedy akademy algorithm to selection a percentage of
    random actions, and a fully-connected neural network for decisions.
    Args:
        action_count: the number of checkpoint_save_dir features in the neural network.
        state_shape: the shape of the input features in the network
        epsilon_min: min value for Epsilon-Greedy exploration. Default: .001
        gamma: discount rate for future rewards in value function. Default: .95
        batch_size: number of previous experiences on which to train. Default: 64
        learning_rate (α): the rate at which model weights are adjusted wrt to loss.
            Default: .001
        hidden_n: the number of nodes in the hidden network layer. Default: 512
        checkpoint_save_dir: filepath to directory in which model weights are
            saved incrementally. Default None saves to current working directory.
        cpu_mode: Optional flag to enable non-GPU training for e.g. non-CUDA
            enabled machines.
    """
    __name__ = "dqn_agent"

    def __init__(
            self,
            action_count: int,
            state_shape: Sequence[int],
            hidden_n: int,
            epsilon_min: float = .001,
            gamma: float = .95,
            batch_size: int = 64,
            learning_rate: float = .001,
            checkpoint_save_dir: str = None,
            cpu_mode: bool = False
    ):
        self.epsilon = 1.0
        self.epsilon_min = epsilon_min
        self.decay_rate = .99998
        self.gamma = gamma
        self.batch_size = batch_size
        self.hidden_n = hidden_n
        self.learning_rate = learning_rate
        self.checkpoint_save_dir = '.' if not checkpoint_save_dir else checkpoint_save_dir
        self.epoch = 0
        self.cpu_mode = cpu_mode
        self.action_count = action_count
        self.state_shape = state_shape

        # actions are discrete
        self.actions = list(range(action_count))

        # defines the memory replay buffer
        self.memory: ExperienceReplayMemory = ExperienceReplayMemory(
            state_shape=self.state_shape
        )

        # defines the neural network
        self.policy_network = ThreeLayerLinear(
            input_n=self.state_shape[0],
            output_n=self.action_count,
            hidden_n=self.hidden_n,
            learning_rate=self.learning_rate,
            cpu_mode=self.cpu_mode
        )

    def get_epsilon(self) -> float:
        """
        Returns the current value of epsilon.
        """
        return self.epsilon

    def get_replay_memory(self) -> ExperienceReplayMemory:
        """
        Returns the replay memory buffer, exposing methods such as sample.
        """
        return self.memory

    def get_name(self) -> str:
        """
        Returns the name of this Agent
        """
        return self.__name__

    def get_action(self, state: np.ndarray) -> int:
        """
        Epsilon-Greedy algorithm to choose random vs. optimal actions using the
        Policy network at all times.
        """
        if random.random() < self.epsilon:
            return random.choice(self.actions)

        state = torch.tensor(state).float().detach()
        state = state.to(self.policy_network.device)
        state = state.unsqueeze(0)
        q_values = self.policy_network(state)
        return torch.argmax(q_values).item()

    def sample_exp(self) -> Tuple[tensor, tensor, tensor, tensor, tensor]:
        """
        Gets a sample of experiences from the buffer and put all to the current
        device as PyTorch tensor objects.
        Returns:
            tuple of tensors representing state, action, reward, next state, dones
        """
        batch_size = self.batch_size
        # in case there are fewer samples than requested batch size
        if len(self.memory) < self.batch_size:
            batch_size = len(self.memory)

        # sample the experience buffer
        states, actions, rewards, next_states, dones = self.memory.sample(
            batch_size=batch_size
        )

        # convert to tensors
        states = torch.tensor(
            states, dtype=torch.float32).to(self.policy_network.device)
        actions = torch.tensor(
            actions, dtype=torch.long).to(self.policy_network.device)
        rewards = torch.tensor(
            rewards, dtype=torch.float32).to(self.policy_network.device)
        next_states = torch.tensor(
            next_states, dtype=torch.float32).to(self.policy_network.device)
        dones = torch.tensor(
            dones, dtype=torch.bool).to(self.policy_network.device)

        return states, actions, rewards, next_states, dones

    def infer(self, state: np.ndarray) -> int:
        """
        Make an inference using the policy network and return the action with
        the maximum reward probability.
        """
        # disable batch normalization and dropouts
        self.policy_network.eval()

        # load w/o backprop and autograd to save memory
        with torch.no_grad():
            state = torch.tensor(state).float().detach()
            state = state.to(self.policy_network.device)
            state = state.unsqueeze(0)
            q_values = self.policy_network(state)
            return torch.argmax(q_values).item()

    def save(self, path: str, delete_old: bool = True) -> True:
        """
        Uses the torch.save function to save a model and delete and
        old ones in the directory specified.
        """
        # save the model
        self.policy_network.save(path=path)

        # delete old checkpoints
        if delete_old:

            # delete files using name as key.
            remove_old_file_versions(filepath=path, remove_key=self.get_name())

        return os.path.exists(path)

    def load(self, *args, **kwargs):
        """
        Wrapper function for network's load function
        """
        return self.policy_network.load(*args, **kwargs)

    def train(self) -> float or None:
        """
        If sufficient experiences exist, sample a batch of previous experiences
        and perform gradient descent to optimize losses via PyTorch's Adam
        optimizer and MSE loss function. Will use the DDQN target/policy network
        design if the class has been instantiated with the self.double param
        indicated as True.
        Returns:
            Calculated loss of this iteration as a float or None if too few
            memories exist in the replay buffer.
        """
        # return if not enough replay experiences to train on yet.
        if len(self.memory) < self.batch_size:
            return None

        # sample from the experience replay buffer
        states, actions, rewards, next_states, dones = self.sample_exp()

        # create an array with indices representing each batch
        # 1D array of shape (batch_size, )
        batch_indices = np.arange(self.batch_size, dtype=np.int64)

        # gets the highest reward actions for this state
        q_values = self.policy_network(states)

        # get predicted values using policy network
        predicted_value_of_now = q_values[batch_indices, actions]

        # get predicted future values using target network if double dqn
        # design is specified, otherwise use policy network
        next_q_values = self.policy_network(next_states)

        # get predicted values of future
        predicted_value_of_future = torch.max(next_q_values, dim=1)[0]

        # discount future rewards based on gamma
        q_target = rewards + self.gamma * predicted_value_of_future * dones

        # calculate loss and perform gradient descent
        loss = self.policy_network.loss(q_target, predicted_value_of_now)
        self.policy_network.optimizer.zero_grad()
        loss.backward()
        self.policy_network.optimizer.step()

        # update action/explore rate - to a minimum
        self.epsilon *= self.decay_rate
        self.epsilon = max(self.epsilon_min, self.epsilon)

        return loss.item()



"""
This file contains code representing a trading environment that is compatible
with OpenAI's Gym (now Gymnasium) framework such that Reinforcement Learning
Agents can train in a standardized fashion.
"""
import logging
from datetime import datetime
from typing import Optional, Union, List, Tuple, NoReturn

import numpy as np
import pandas as pd
from gymnasium.core import RenderFrame, ObsType
from gymnasium.spaces import Discrete, Box

from akademy.common.utils import format_float
from akademy.models.base_models.trade_env_base import TradeEnvBase
from akademy.models.trade import Trade
from akademy.models.trade_action import TradeAction
from akademy.common.utils import minmax_normalize


class TradeEnv(TradeEnvBase):
    def __init__(
            self,
            data: pd.DataFrame,
            window: int,
            asset: str,
            starting_cash: float = 100000,
            pct_based_rewards: bool = True,
            fractional: bool = False,
    ):
        """
        Trading Environment in which simulated trading can occur. Data should
        be a dataframe with the Open, High, Low, Close and Volume dimensions
        available and indexed by date.

        Args:
            data: pandas DataFrame object, indexed by date, with OHLCV dimensions
            window: the number of past periods of which observations are aware.
            starting_cash: amount of starting cash in the account.
            asset: name of the asset being traded e.g. SPY.
            pct_based_rewards: if disabled, will calculate rewards as absolute values
                rather than % based comparisons to Buy/Hold strategies.
            fractional: option to allow non-integer purchases of assets like e.g. BTC
        """
        # show some info
        logging.info(f"TradeEnv instantiated with {len(data)} rows of data.")

        # store reference to raw data + create initial sample
        self.window = window
        self.data = data

        # validate the data
        self._validate_data()

        # configure environment
        self.asset = asset
        self.starting_cash = starting_cash
        self.cash = self.starting_cash
        self.i = self.window
        self.qty = 0
        self.trades = []

        # the amount at which the account has, functionally, zero capital left.
        self._zero = min(self.starting_cash * .0001, 10)

        # amounts of fractional holdings where a sale isn't likely to be possible.
        self._dust = self._zero / self.data['close'].mean()

        # if true, allows fractional purchase of assets (e.g. .001 BTC)
        self.fractional = fractional
        self.pct_based_rewards = pct_based_rewards

        # define space dynamics.
        self.action_space = Discrete(3)

        # define the space as a series of observations, each normalized to
        # a range of [0, 1] where the total number of each is calculated as
        # < self.window * 5 + 1 > where the <window> is each historical price data,
        # the 5 is the OHLCV for each period, and the <1> is the single opening
        # price of the current period.
        low = np.array(
            [0] * (self.window * 5 + 1)
        )
        high = np.array(
            [1] * (self.window * 5 + 1)
        )
        self.observation_space = Box(
            low=np.float32(low),
            high=np.float32(high),
            dtype=np.float32
        )

        # get buy/hold performance data
        self.buy_hold_abs, \
        self.buy_hold_pct = self._buy_hold_performance_data()
        self.buy_hold_qty = self._bh_qty()

    @property
    def total_equity(self) -> float:
        """
        Calculates the total equity as the sum of all current cash plus the
        market value of total assets held at the current trading period's open.
        """
        return self.cash + (self.qty * self.data.iloc[self.i]['open'])

    def get_trades(self) -> List[Trade]:
        """
        Returns the Trade record from this environment as a list of Trade class
        objects.
        """
        return self.trades

    def _bh_qty(self):
        """
        Calculates the initial qty for the B/H strategy.
        """
        qty = self.starting_cash / self.data.iloc[0]['open']
        if not self.fractional:
            qty = int(qty)
        return qty

    def _validate_data(self):
        """
        Runs a series of validation checks on the ENV data to ensure the following:
            1. No NaN entries present.
        """
        # check no NaN values in the DataFrame
        nans = self.data.isna().sum().sum()
        if nans != 0:
            raise Exception(
                f"TradeEnv data contains {nans} NaN values."
            )

    def _buy_hold_performance_data(self) -> Tuple[float, float]:
        """
        Calculates Buy/Hold stats for the current trading data starting
        with the open price of the first trade period and ending with the
        close of the last trade period -- somewhat unrealistic.
        Returns:
            Tuple of B/H performance data as pnlAbs, pnlPct
        """
        qty = self.starting_cash / self.data.iloc[0]['open']
        if not self.fractional:
            qty = int(qty)
        end_equity = qty * self.data.iloc[-1]['close']
        return end_equity - self.starting_cash, ((end_equity - self.starting_cash) / self.starting_cash) * 100

    def _get_info(self) -> dict:
        """
        Current state of the environment
        """
        return {
            "i": self.i,
            "cash": f"${format_float(self.cash)}",
            "qty": self.qty,
            "pnlAbs": f"${format_float(self._pnl_abs())}",
            "pnlPct": f"{format_float(self._pnl_pct())}%",
            "totalEquity": f"${format_float(self.total_equity)}",
        }

    def _hodl_total_equity(self):
        """
        Gets the total equity of the Buy/Hold strategy with consideration for
        fractional trading if enabled.
        """
        qty = (self.starting_cash / self.data.iloc[0]['open'])
        if not self.fractional:
            qty = int(qty)
        return qty * self.data.iloc[-1]['close']

    def _pnl_pct(self):
        """Calculates current pnl as pct change"""
        return (self.total_equity - self.starting_cash) / self.starting_cash * 100

    def _pnl_abs(self):
        """Calculates current pnl in cash"""
        return self.total_equity - self.starting_cash

    def _is_terminal(self):
        """
        Checks the environment to see if a terminal state is reached, indicating
        that one of the following conditions are met:
            1. no more trading data
            2. no cash and not holding

        Any of this conditions result in a False status that will indicate
        no further trading should take place in the environment.

        Returns:
            Boolean indicating terminal status
        """
        done = False
        if self.i + 1 >= len(self.data):
            logging.info("Terminal State Reached: self.i + 1 >= len(self.data)")
            done = True
        if self.cash <= self._zero and self.qty <= self._dust:
            logging.info("Terminal State Reached: self.cash <= self._zero AND self.qty <= self._dust")
            done = True
        return done

    def _make_observation(self) -> pd.DataFrame:
        """
        Fills the observation window up with new state information.
        Note:
            normalizes observation data to range of (0, 1) on a per-window basis
        Returns:
            the current NP array.
        """
        return self.data.iloc[self.i - self.window: self.i + 1]

    def _sample_price(self) -> float:
        """
        Sample a price from the current periods' range of [low, high]
        to simulate a purchase at an unknown intra-period timestamp.
        """
        _low = self.data.iloc[self.i]['low']
        _high = self.data.iloc[self.i]['high']

        return np.random.uniform(low=_low, high=_high, size=1)[0]

    def _make_report(self, price: float, qty: float, date: datetime, side: int) -> Trade:
        """
        Creates a trade object, saves to local object memory, returns for reference
        """
        report = Trade(
            price=price,
            date=date,
            asset=self.asset,
            qty=qty,
            side=side
        )
        self.trades.append(report)
        return report

    def _buy(self) -> NoReturn:
        """
        Applies a BUY action to the environment.
        """
        # if the account is busted return
        if self.cash <= self._zero:
            return

        # get a sampled price and calculate position size
        # Note: Full allocations are made each time.
        _price = self._sample_price()
        _qty = self.cash / _price
        if not self.fractional:
            _qty = int(_qty)

        # make this a terminal status so this is never reached
        # exception is a reminder
        if _qty == 0 and self.qty == 0:
            print("Cash:", self.cash, "QTY:", self.qty)
            raise Exception(f'TradeEnv is Broke')

        # make the purchase
        self.cash -= (_qty * _price)
        self.qty += _qty

        # make a trade report
        report = self._make_report(
            price=_price,
            qty=_qty,
            date=self.data.index[self.i].to_pydatetime(),
            side=TradeAction.BUY.value
        )
        logging.debug(f"BUY Action: {report}")

    def _sell(self):
        """Applies a SELL action to the environment"""
        # if not holding, can't sell
        if self.qty <= .01:
            return

        # get randomly
        _price = self._sample_price()

        # make sell
        self.cash += self.qty * _price
        _qty = self.qty
        self.qty = 0

        # make report -- appends to trades
        report = self._make_report(
            price=_price,
            qty=_qty,
            date=self.data.index[self.i].to_pydatetime(),
            side=TradeAction.SELL.value
        )
        logging.debug(f'SELL: {report}')

    def _hold(self):
        """
        For book-keeping purposes, to track the total number of
        actions. Hold actions don't really do anything that would
        need a record, in the Buy/Sell context.
        """
        _price = self.data.iloc[self.i]['open']

        # make report -- appends to trades
        report = self._make_report(
            price=_price,
            qty=self.qty,
            date=self.data.index[self.i].to_pydatetime(),
            side=TradeAction.HOLD.value
        )
        logging.debug(f'HOLD: {report}')

    def _take_action(self, action: int):
        """Applies an action to the environment"""
        if action == TradeAction.BUY.value:
            return self._buy()
        elif action == TradeAction.SELL.value:
            return self._sell()
        elif action == TradeAction.HOLD.value:
            return self._hold()

        # shouldn't ever reach this
        else:
            raise Exception(f"Invalid Action: {action}")

    def step(self, action: int) -> Tuple[np.array, float, bool, bool, dict, bool]:
        """
        Applies an action to the current environment and returns an observation
        of the resulting state.
        Args:
            action: the action

        Returns:
            Tuple as (next_state, reward, terminality, False, info, done)
        """
        # equity before any trade action is taken
        _trade_initial_equity = self.total_equity
        _bh_initial_equity = self.data.iloc[self.i]['open'] * self.buy_hold_qty

        # take action
        self._take_action(action)

        # bumps the observation period BEFORE calculating rewards as a way
        # to ensure rewards reflect the actions of a period, realized afterwards.
        self.i += 1

        # calculate %-based rewards or absolute rewards based on config.
        if self.pct_based_rewards:

            # get trade pct change
            trade_pct = (self.total_equity - _trade_initial_equity) / _trade_initial_equity * 100

            # get hodl pct change
            _bh_final_equity = self.data.iloc[self.i]['open'] * self.buy_hold_qty
            hodl_pct = (_bh_final_equity - _bh_initial_equity) / _bh_initial_equity * 100

            # reward is difference
            reward = trade_pct - hodl_pct

        else:
            reward = self.total_equity - _trade_initial_equity

        # as next_state, reward, terminality, info
        terminal = self._is_terminal()
        return self._make_observation(), reward, terminal, False, self._get_info(), terminal

    def reset(
            self,
            *,
            seed: Optional[int] = None,
            options: Optional[dict] = None,
    ) -> Tuple[ObsType, dict]:
        """
        Function to reset the trading environment to the initial state between
        trading episodes. Takes the following actions:
            1. resets the current period based on lookback window value
            2. resets the amount of cash to the starting balance
            3. resets the value of qty to 0
            4. removes all trade records
        Returns:
            A numpy array containing a single observation of the initial state + info dict
        """

        self.i = self.window
        self.cash = self.starting_cash
        self.qty = 0
        self.trades = []

        # reset data-dependent features
        self.buy_hold_qty = self._bh_qty()
        self.buy_hold_abs,\
        self.buy_hold_pct = self._buy_hold_performance_data()

        return self._make_observation(), self._get_info()

    def render(self) -> Optional[Union[RenderFrame, List[RenderFrame]]]:
        """Unused"""
        return self._get_info()



"""
This file contains a listing of different type definitions used throughout
the project.
"""
from typing import TypeVar

# input features to Neural Networks
State = TypeVar("State")

# values of checkpoint_save_dir features from Neural Networks (discrete or continuous)
Action = TypeVar("Action", int, float)



import random
from unittest import TestCase

from akademy.common.utils import load_spy_daily
from akademy.models.envs.trade_env import TradeEnv


class TestTradeEnv(TestCase):
    """
    Tests various aspects of the TradeEnv.
    """
    def setUp(self) -> None:

        self.asset = "btc"
        self.window = 50
        self.frequency = "1d"
        self.count = self.window * 3
        self.data = load_spy_daily(count=self.count)
        self.env = self._make_trade_env()

    def _make_trade_env(self):
        """
        Makes a trade environment based on setup params.
        """
        return TradeEnv(
            data=self.data,
            window=self.window,
            asset=self.asset,
            pct_based_rewards=True,
            fractional=True
        )

    def test_step(self):
        """
        Steps the environment many times to ensure all is well.
        """
        # simulate 100 episodes where an agent might progress through
        # the environment and ensure the environment can increment and reset.
        count = 0
        episodes = 25
        for i in range(episodes + 1):
            count = i
            self.env.reset()
            done = False
            while not done:

                # iterate state
                state, reward, done, truncated, info, _deprecated = self.env.step(
                    action=random.choice([0, 1, 2])
                )
        self.assertEqual(count, episodes)

        # reset for next use
        self.env.reset()

    def test_internal_data_length(self):
        """
        Test that the environment contains the expected amount of data.
        """
        self.assertTrue(len(self.env.data) == self.count)

    def test_make_observation(self):
        """
        Observations should be <window_size * 5 (OHLCV) + 1 (current_open)>
        """
        self.assertTrue(self.env.window == self.window)
        self.assertTrue(
            len(self.env._make_observation()) == self.window * 5 + 1
        )



from abc import ABC, abstractmethod


class EpsilonGreedy(ABC):
    """
    Class that requires concretions to maintain various properties associated
    with an Epsilon-Greedy Policy as described in the following paper:
        Mnih, V., Kavukcuoglu, K., Silver, D. et al.
        Human-level control through deep reinforcement learning.
        Nature 518, 529–533 (2015). https://doi.org/10.1038/nature14236
    """
    @abstractmethod
    def get_epsilon(self) -> float:
        """
        Returns an experience replay buffer object from which an agent can
        sample previous experiences.
        """
        raise NotImplementedError



import logging

import torch
import torch.nn as nn
from torch.nn import MSELoss
from torch.optim import Adam

from akademy.models.base_models.network_base import NetworkBase


class ThreeLayerLinear(NetworkBase):
    """
    A three-layer linear (a.k.a. fully connected) neural network that uses the
    Adam optimizer, MSE loss function, and ReLU activation functions between
    layers.
    Args:
        input_n: the number of input features to the network.
        output_n: the number of checkpoint_save_dir features from the last layer.
        hidden_n: the size of the hidden layer features.
        learning_rate: the learning rate at which the optimizer is initialized.
        cpu_mode: optional flag to force CPU mode.
    """
    def __init__(self,
                 input_n: int,
                 output_n: int,
                 hidden_n: int,
                 learning_rate: float,
                 cpu_mode: bool = False

                 ):
        super().__init__()

        # determine on which device the model will be located
        if not cpu_mode:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            self.device = "cpu"

        # define network dimensions for easy reference
        self.input_n = input_n
        self.output_n = output_n
        self.hidden_n = hidden_n

        # store reference to the learning rate value
        self.learning_rate = learning_rate

        # define the model using PyTorch's Sequential class
        self.model = nn.Sequential(
            nn.Linear(in_features=self.input_n, out_features=self.hidden_n),
            nn.ReLU(),
            nn.Linear(in_features=self.hidden_n, out_features=self.hidden_n),
            nn.ReLU(),
            nn.Linear(in_features=self.hidden_n, out_features=self.output_n)
        )

        # define the optimizer and loss calculations
        self.optimizer = Adam(params=self.model.parameters(), lr=self.learning_rate)
        self.loss = MSELoss()

        # put the model to the device
        self.model.to(self.device)

    def load(self, path: str) -> bool:
        """
        Loads the weights from a checkpoint file into the current network.
        """
        logging.info(f'Loading pretrained model: {path}')
        self.model.load_state_dict(torch.load(path))
        return True

    def save(self, path: str):
        """
        Saves the current network configuration in a file.
        """
        logging.info(f'Saving current model to: {path}')
        torch.save(self.model.state_dict(), path)

    def forward(self, x):
        """
        This method is included here for backwards compatability but is no
        longer strictly necessary as the model call can be made directly.
        """
        return self.model(x)



import os.path
from unittest import TestCase
from akademy.common.project_paths import SPY_DATA
import pandas as pd


class TestData(TestCase):
    """
    Test suite to ensure all expected data is available to the project.
    """
    def test_spy_csv(self):
        """
        Tests that the SPY.csv file is available as expected.
        """
        self.assertTrue(os.path.exists(SPY_DATA))
        df = pd.read_csv(SPY_DATA)
        self.assertTrue(all([

            # test all expected columns required for training are present
            [
                x in df.columns
                for x in ['Date', 'Open', 'High', 'Low', 'Close', 'Volume']
            ],

            # test the expected length
            len(df) > 7454

        ]))



from abc import abstractmethod, ABC
from typing import Any, NoReturn, Iterable

from akademy.common.typedefs import State, Action


class ReplayMemory(ABC):
    """
    Base class for replay buffers that ensure essential methods are available.
    """
    @abstractmethod
    def add(self,
            state: State, action: Action, reward: Any,
            next_state: State, done: bool) -> NoReturn:
        """
        Adds a single experience to the memory buffer.
        """
        raise NotImplementedError

    @abstractmethod
    def sample(self, batch_size: int) -> Iterable:
        """
        Returns a collection of samples from the replay buffer.
        """
        raise NotImplementedError



import torch.nn as nn
from abc import abstractmethod


class NetworkBase(nn.Module):
    """
    Base class for all networks that adds a load/save requirement for torch's
    Module class.
    """
    @abstractmethod
    def load(self, path: str):
        """
        Load weights into the model from a local file.
        """
        raise NotImplementedError

    @abstractmethod
    def save(self, path: str):
        """
        Save network state to a local filepath.
        """
        raise NotImplementedError



from datetime import datetime
import json
from typing import NoReturn

from akademy.models.trade_action import TradeAction


class Trade:
    """
    An easily serializable report of trades made during an episode.
    """
    def __init__(
            self, price: float, date: datetime,
            asset: str, qty: float, side: int
    ) -> NoReturn:
        self.price = price
        self.date = date
        self.asset = asset
        self.qty = qty
        self.side = side

    def __dict__(self) -> dict:
        """
        Converts the Trade object to a dictionary representation where the
        only conversion is the <side> field such that the TradeAction object
        is converted to an explicit str value of either "BUY" or "SELL" for
        clarity.
        """
        return {
            "price": float(self.price),
            "date": str(self.date.isoformat()),
            "asset": self.asset,
            "qty": self.qty,
            "side": "BUY" if self.side == TradeAction.BUY.value else "SELL"
        }

    def to_json(self) -> str:
        """
        Converts the Trade object into a JSON-formatted str representation such
        that all numerical values are strings and the TradeAction objects in the
        <side> field are converted to "BUY" or "SELL" for clarity.
        """
        return json.dumps(self.__dict__())

    def __str__(self) -> str:
        """
        Wraps the <to_json> method for object str representation.
        """
        return self.to_json()



import json
from typing import NoReturn


class OHLCV:
    """
    Object model for a single period's pricing history as
        open, low, high, close, volume and date
    Note:
        Expects all values to be numeric. Convert if using raw CSV data.
        <date> object is expected to be a Unix-format timestamp.
    """
    def __init__(
            self, date: int, open: float, high: float,
            low: float, close: float, volume: int
    ) -> NoReturn:
        self.date = date
        self.open = float(open)
        self.high = float(high)
        self.low = float(low)
        self.close = float(close)
        self.volume = int(volume)

    def __dict__(self) -> dict:
        """
        Dictionary representation of the object.
        """
        return {
            'date': self.date,
            'open': self.open,
            'high': self.high,
            'low': self.low,
            'close': self.close,
            'volume': self.volume,
        }

    def to_json(self) -> str:
        """
        JSON-formatted string representation of the object.
        """
        return json.dumps(self.__dict__())

    def __str__(self) -> str:
        """
        Display as JSON-formatted str.
        """
        return self.to_json()



from abc import ABC, abstractmethod
from typing import List

from gymnasium import Env

from akademy.models.trade import Trade


class TradeEnvBase(Env, ABC):
    """
    Base class for all Trading Environments that extends the Gym ENV to
    offer a few extra methods.
    """
    @property
    def total_equity(self) -> float:
        """
        Gets the total equity of the environment in its current state.
        """
        raise NotImplementedError

    @abstractmethod
    def get_trades(self) -> List[Trade]:
        """
        Returns a List of Trade objects representing the trades made within
        an Environment object.
        """
        raise NotImplementedError




from typing import Tuple, NoReturn, Sequence

import numpy as np
from numpy import ndarray

from akademy.models.base_models.replay_memory import ReplayMemory
from akademy.common.typedefs import State, Action


class ExperienceReplayMemory(ReplayMemory):
    """
    Implementation of the Experience Replay model as first described in the
    following paper:
        Lin, LJ. Self-improving reactive agents based on reinforcement learning,
        planning and teaching. Mach Learn 8, 293–321 (1992).
        https://doi.org/10.1007/BF00992699
    Args:
        state_shape: Tuple of a Tensor's dimensions e.g. (251, ) for 50 window
        memsize: total number of possible experiences to hold
    """
    def __init__(self, state_shape: Sequence[int], memsize: int = 10000):
        self.mem_count = 0
        self.memsize = memsize
        self.state_shape = state_shape

        # allocate numpy arrays as buffers to hold replay experiences
        self.states = np.zeros((self.memsize, *self.state_shape), dtype=np.float32)
        self.actions = np.zeros(self.memsize, dtype=np.int64)
        self.rewards = np.zeros(self.memsize, dtype=np.float32)
        self.next_states = np.zeros((self.memsize, *self.state_shape), dtype=np.float32)
        self.dones = np.zeros(self.memsize, dtype=bool)

    def add(
            self, state: State, action: Action,
            reward: float, next_state: State, done: bool
    ) -> NoReturn:
        """
        Add a replay experience as constituent parts to the appropriate buffers
        using the mem_index to treat them like ring buffers.
        Args:
            state: Object e.g. ndarray representing the current state of an environment.
            action: Object e.g. float | int representing action taken by Agent.
            reward: Value of the reward generated by the Agent's action.
            next_state: Object e.g. ndarray representing the resulting state of an environment
                after an Agent's action is applied.
            done: boolean indicating if this was a terminal state.
        """
        # treat buffers like FIFO stacks
        mem_index = self.mem_count % self.memsize

        # adds each part of the experience into the appropriate index of
        # the corresponding ndarrays.
        self.states[mem_index] = state
        self.actions[mem_index] = action
        self.rewards[mem_index] = reward
        self.next_states[mem_index] = next_state
        self.dones[mem_index] = 1 - done  # flip the bit here

        # keeps a measure of each addition such that modulo operations
        # can be used to effectively determine the last-added reference
        # such that the memory acts as a FIFO queue.
        self.mem_count += 1

    def sample(
            self, batch_size: int = 64
    ) -> Tuple[ndarray, ndarray, ndarray, ndarray, ndarray]:
        """
        Gets a list of <batch_size> many random indices from within
        a range of [0, min(memcount, memsize)] and returns the elements from
        each ring buffer corresponding to those as a tuple as:

        Args:
            batch_size: number of total random samples to return.

        Returns:
            Tuple of ndarrays as [state, action, reward, next-states, dones]
        """
        # control bounds of sample size to ensure
        #   A.) no more than max amount sampled
        #   B.) no more tha min available sampled.
        _mem_max = min(self.mem_count, self.memsize)

        # randomly select indices to be sampled
        batch_indices = np.random.choice(_mem_max, batch_size, replace=True)

        # create sub-array from each specified index, from each component.
        states = self.states[batch_indices]
        actions = self.actions[batch_indices]
        rewards = self.rewards[batch_indices]
        next_states = self.next_states[batch_indices]
        dones = self.dones[batch_indices]

        return states, actions, rewards, next_states, dones

    def __len__(self) -> int:
        """
        Length is reported as the minimum of the count of added experiences and
        the maximum number of experiences such that the total possible range
        of length values is [0, memsize].
        Returns:
            integer value representing the current capacity of the
            ExperienceReplay object in the range of (0, memsize]
        """
        return min(self.mem_count, self.memsize)



from abc import ABC, abstractmethod

from akademy.common.typedefs import State, Action


class Agent(ABC):
    """
    Base class for all learning agents
    """
    @abstractmethod
    def train(self):
        """
        Method to initialize training routine for the agent.
        """
        raise NotImplementedError

    @abstractmethod
    def infer(self, state: State) -> Action:
        """
        Given an observation of state load network for inference.
        """
        raise NotImplementedError

    @abstractmethod
    def get_name(self) -> str:
        """
        Returns a name of a particular agent. Useful for debugging, and file
        saving.
        """
        raise NotImplementedError

    @abstractmethod
    def get_action(self, state: State) -> Action:
        """
        Gets an action without loading enabling eval() mode or no_grad().
        """
        raise NotImplementedError



from akademy.models.experience_replay_memory import ExperienceReplayMemory
from akademy.models.ohlcv import OHLCV
from akademy.models.trade import Trade
from akademy.models.trade_action import TradeAction



from akademy.models.base_models.agent import Agent
from akademy.models.base_models.epsilon_greedy import EpsilonGreedy
from akademy.models.base_models.network_base import NetworkBase
from akademy.models.base_models.replay_memory import ReplayMemory
from akademy.models.base_models.trade_env_base import TradeEnvBase



import os
from setuptools import setup


def get_readme():
    """
    Convenience function to read in the contents of the README.md file
    such that it can be conveniently included in the "long_description".
    """
    return open(os.path.join(os.path.dirname(__file__), 'README.md')).read()


setup(
    name="akademy",
    version="0.1.42",
    url="https://github.com/alphazwest/akademy",
    author="Zack West",
    author_email="alphazwest@gmail.com",
    description="akademy: A Reinforcement Learning Framework",
    license="BSD-3-Clause",
    keywords=[
        "reinforcement learning", "quantitative trading", "fintech",
        "trading bot", "algorithmic trading", "finance", "automated trading",
        "neural networks", "artificial intelligence", "machine learning"
    ],
    packages=[
        'tests',
        'akademy',
        'akademy.common',
        'akademy.models',
        'akademy.models.agents',
        'akademy.models.envs',
        'akademy.models.base_models',
        'akademy.models.networks'
    ],
    long_description=get_readme(),

    # see here: https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",

        # optional GPU usage, but 11.7 CUDA required for GPU via PyTorch
        "Environment :: GPU :: NVIDIA CUDA :: 11.7"
    ],
    # note: PyTorch allows 3.7 for Windows, but requires >=3.7.6 for Linux
    #       but formally recommends >=3.8.1 when building from source.
    #       see here: https://pypi.org/project/torch/ (check version == 1.13.1)
    python_requires=">=3.7",
    install_requires=[
        "pandas==1.5.2",
        "torch==1.13.1",
        "gymnasium==0.27.0"
    ],
    provides=["akademy"]
)



import logging
import os
from typing import List

import numpy as np
import pandas as pd
from akademy.common import project_paths


def format_float(value: float, do_rounding: bool = True) -> str:
    """
    Convenience function to adjust report checkpoint_save_dir formatting.
    Turns floating point numbers into comma-separated strings
    rounded to the nearest 2 decimal places.
    """
    if do_rounding:
        value = round(value, 2)
    return f"{value:,}"


def get_file_extension(filepath: str) -> str:
    """
    Given the path to a file, return only the extension such that the following
    cases are handled:
        1. file.ext -> ext
        2. file.tar.gz -> tar.gz
        3. file -> ''
    """
    filename = os.path.basename(filepath)
    if "." not in filename:
        return filename
    return ".".join(filename.split('.')[1:])


def remove_old_file_versions(
        filepath: str,
        remove_key: str,
        keep_key: str = None,
        keep_count: int = 0
) -> List[str]:
    """
    Given an absolute filepath to a file, removes all files within that directory
    that contain the <key> value within them, keeping <keep> many files,
    where the deleted files are those oldest via last_mod check.
    Args:
        filepath: path to the file from which to remove old versions.
        keep_count: the number of files to keep. Default 0 means delete them all.
        remove_key: character sequence to identify similar files.
            e.g. if <key> in <file>
        keep_key: files matching <keep_key> in <file> are not considered.
    Returns:
        List of deleted filepaths.
    """
    file_dir = os.path.abspath(os.path.dirname(filepath))
    file_name = os.path.basename(file_dir).split('.')[0]
    file_ext = ".".join(os.path.basename(filepath).split('.')[1:])

    # get list of filenames intended for removal.
    files = [f for f in os.listdir(file_dir) if remove_key in f]

    # remove files matching keep mask
    if remove_key:
        files = [f for f in files if keep_key not in f]

    # sort by filemod time
    files = sorted(files,
                   key=lambda x: os.path.getmtime(
                       os.path.abspath(os.path.join(file_dir, x))
                   ),
                   reverse=True
                   )

    # keep max if specified
    if keep_count > 0:
        if keep_count > len(files):
            pass
        else:
            files = files[keep_count:]

    # delete the older files
    deleted = []
    for file in files:

        fullpath = os.path.abspath(os.path.join(
            file_dir,
            file
        ))
        deleted.append(fullpath)
        logging.info(f'removing file: {fullpath}')
        os.remove(fullpath)

    return deleted


def load_csv_ohlcv_data(
        filepath: str,
        count: int = 0,
        date_ms_unit: str = "ns") -> pd.DataFrame:
    """
    Loads financial data presumed to have the following columns present:
        Open, High, Low, Close, Volume, Date
    Args:
        filepath: the path to the CSV file to open.
        count: total number of records to return; earlier rows dropped; 0 == all
        date_ms_unit: the unit measure of the date timestamp. Default Nanoseconds.
    Return:
        data as a DataFrame with open, high, low, close, volume
        headers and indexed by ascending date
    """
    data = pd.read_csv(filepath)

    # lowercase all column names for standardization
    data.columns = [x.lower() for x in data.columns]

    # drop all but essential pricing + date + volume
    data = data[['date', 'open', 'high', 'low', 'close', 'volume']]

    # convert types
    data = data.astype(
        {
            'date': f'datetime64[{date_ms_unit}]',
            'open': np.float32,
            'high': np.float32,
            'low': np.float32,
            'close': np.float32,
            'volume': np.float32
        }
    )

    # reindex
    data.set_index('date', inplace=True)

    # 0 == all the data
    if count == 0:
        return data

    # check more data not asked for than available
    if count > len(data):
        raise Exception(
            f"The request number of records <{count}> "
            f"is more than total available: <{len(data)}>"
        )

    # otherwise, return the <count> most recent data items
    return data[len(data) - count:]


def load_spy_daily(count: int = 0):
    """
    Loads the locally-saved SPY pricing data with daily OHLCV prices
    ranging from 1/29/1993 to 9/6/2022 in ascending order.
    7455 total entries in daily increments.
    Arguments:
        count: number of days to return from the max period. End is most recent.
    Notes:
        'close' is replaced by 'Adj Close'; but renamed as 'close'
    Returns:
        pd.DataFrame object
    """
    return load_csv_ohlcv_data(project_paths.SPY_DATA, count=count)


def flatten_price_obs(obs: pd.DataFrame) -> np.ndarry:
    """
    Given a DataFrame (or slice) of
    """
    if not all(c in obs.columns for c in
               ['open', 'high', 'low', 'close', 'volume']):
        raise Exception("Price observations must contain OHLCV columns.")
    # converts to NumPy array, flattens the data, and clips all but the opening
    # price of the last column (current known price)
    return obs.to_numpy().flatten()[:-4]


def minmax_normalize(data: np.ndarray) -> np.ndarray:
    """
    Converts all values within an ndarray to values within the range of 0-1
    where
    Args:
        data: numpy array of data.
    Returns:
        a new numpy array containing values in range 0-1
    """
    _min = np.min(data)
    _ptp = np.ptp(data)
    return (data-_min)/_ptp
